@@ -1,4 +1,4 @@
-use super::{AcceptFlags, InitiateFlags, K5Ctx, K5ServerCtx};
+use super::{AcceptFlags, InitiateFlags, K5Ctx, K5ServerCtx, Step};
 use anyhow::{anyhow, bail, Result};
 use bytes::{buf::Chain, Buf, BytesMut};
 use std::{
@@ -10,35 +10,37 @@ use std::{
     ptr,
     time::Duration,
 };
-use windows::Win32::{
-    Foundation::{
-        GetLastError, FILETIME, PWSTR, SEC_E_OK, SEC_I_COMPLETE_AND_CONTINUE,
-        SEC_I_COMPLETE_NEEDED, SYSTEMTIME,
-    },
-    Globalization::lstrlenW,
-    Security::{
-        Authentication::Identity::{
-            AcceptSecurityContext, AcquireCredentialsHandleW, CompleteAuthToken,
-            DecryptMessage, DeleteSecurityContext, EncryptMessage, FreeContextBuffer,
-            FreeCredentialsHandle, InitializeSecurityContextW, QueryContextAttributesW,
-            QueryCredentialsAttributesW, QuerySecurityPackageInfoW, SecBuffer,
-            SecBufferDesc, SecPkgContext_NativeNamesW, SecPkgContext_Sizes,
-            SecPkgCredentials_NamesW, SecPkgInfoW, ACCEPT_SECURITY_CONTEXT_CONTEXT_REQ,
-            ISC_REQ_CONFIDENTIALITY, ISC_REQ_MUTUAL_AUTH, KERB_WRAP_NO_ENCRYPT,
-            SECBUFFER_DATA, SECBUFFER_PADDING, SECBUFFER_STREAM, SECBUFFER_TOKEN,
-            SECBUFFER_VERSION, SECPKG_ATTR_NATIVE_NAMES, SECPKG_ATTR_SIZES,
-            SECPKG_CRED_ATTR_NAMES, SECPKG_CRED_INBOUND, SECPKG_CRED_OUTBOUND,
-            SECURITY_NATIVE_DREP,
+use windows::{
+    core::{PCWSTR, PWSTR},
+    Win32::{
+        Foundation::{GetLastError, FILETIME, SEC_E_OK, SYSTEMTIME},
+        Globalization::lstrlenW,
+        Security::{
+            Authentication::Identity::{
+                AcceptSecurityContext, AcquireCredentialsHandleW, DecryptMessage,
+                DeleteSecurityContext, EncryptMessage, FreeContextBuffer,
+                FreeCredentialsHandle, InitializeSecurityContextW,
+                QueryContextAttributesW, QueryCredentialsAttributesW,
+                QuerySecurityPackageInfoW, SecBuffer, SecBufferDesc,
+                SecPkgContext_NativeNamesW, SecPkgContext_Sizes,
+                SecPkgCredentials_NamesW, SecPkgInfoW,
+                ACCEPT_SECURITY_CONTEXT_CONTEXT_REQ, ISC_REQ_CONFIDENTIALITY,
+                ISC_REQ_MUTUAL_AUTH, KERB_WRAP_NO_ENCRYPT, SECBUFFER_CHANNEL_BINDINGS,
+                SECBUFFER_DATA, SECBUFFER_PADDING, SECBUFFER_STREAM, SECBUFFER_TOKEN,
+                SECBUFFER_VERSION, SECPKG_ATTR_NATIVE_NAMES, SECPKG_ATTR_SIZES,
+                SECPKG_CRED_ATTR_NAMES, SECPKG_CRED_INBOUND, SECPKG_CRED_OUTBOUND,
+                SECURITY_NATIVE_DREP, SEC_CHANNEL_BINDINGS,
+            },
+            Credentials::SecHandle,
         },
-        Credentials::SecHandle,
-    },
-    System::{
-        Diagnostics::Debug::{
-            FormatMessageW, FORMAT_MESSAGE_FROM_SYSTEM, FORMAT_MESSAGE_IGNORE_INSERTS,
+        System::{
+            Diagnostics::Debug::{
+                FormatMessageW, FORMAT_MESSAGE_FROM_SYSTEM, FORMAT_MESSAGE_IGNORE_INSERTS,
+            },
+            SystemInformation::GetSystemTime,
+            SystemServices::{LANG_NEUTRAL, SUBLANG_NEUTRAL},
+            Time::{SystemTimeToFileTime, SystemTimeToTzSpecificLocalTime},
         },
-        SystemInformation::GetSystemTime,
-        SystemServices::{LANG_NEUTRAL, SUBLANG_NEUTRAL},
-        Time::{SystemTimeToFileTime, SystemTimeToTzSpecificLocalTime},
     },
 };
 
@@ -47,7 +49,7 @@ fn failed(res: i32) -> bool {
 }
 
 unsafe fn string_from_wstr(s: *mut u16) -> String {
-    let slen = lstrlenW(PWSTR(s));
+    let slen = lstrlenW(PCWSTR(s));
     let slice = &*ptr::slice_from_raw_parts(s, slen as usize);
     OsString::from_wide(slice).to_string_lossy().to_string()
 }
@@ -80,7 +82,7 @@ struct Cred(SecHandle);
 impl Drop for Cred {
     fn drop(&mut self) {
         unsafe {
-            FreeCredentialsHandle(&mut self.0 as *mut _);
+            FreeCredentialsHandle(&mut self.0);
         }
     }
 }
@@ -91,20 +93,20 @@ impl Cred {
         let principal = principal.map(str_to_wstr);
         let principal_ptr =
             principal.map(|mut p| p.as_mut_ptr()).unwrap_or(ptr::null_mut());
-        let mut package = str_to_wstr(if negotiate { "Negotiate" } else { "Kerberos" });
+        let package = str_to_wstr(if negotiate { "Negotiate" } else { "Kerberos" });
         let dir = if server { SECPKG_CRED_INBOUND } else { SECPKG_CRED_OUTBOUND };
         let mut lifetime = 0i64;
         let res = unsafe {
             AcquireCredentialsHandleW(
-                PWSTR(principal_ptr),
-                PWSTR(package.as_mut_ptr()),
+                PCWSTR(principal_ptr),
+                PCWSTR(package.as_ptr()),
                 dir,
                 ptr::null_mut(),
                 ptr::null_mut(),
                 None,
                 ptr::null_mut(),
-                &mut cred as *mut _,
-                &mut lifetime as *mut _,
+                &mut cred,
+                &mut lifetime,
             )
         };
         if failed(res) {
@@ -118,7 +120,7 @@ impl Cred {
         let mut names = SecPkgCredentials_NamesW::default();
         unsafe {
             let res = QueryCredentialsAttributesW(
-                &mut self.0 as *mut _,
+                &mut self.0,
                 SECPKG_CRED_ATTR_NAMES,
                 &mut names as *mut _ as *mut c_void,
             );
@@ -133,19 +135,18 @@ impl Cred {
 fn alloc_krb5_buf() -> Result<Vec<u8>> {
     let mut ifo = ptr::null_mut::<SecPkgInfoW>();
     let mut pkg = str_to_wstr("Kerberos");
-    let res =
-        unsafe { QuerySecurityPackageInfoW(PWSTR(pkg.as_mut_ptr()), &mut ifo as *mut _) };
+    let res = unsafe { QuerySecurityPackageInfoW(PCWSTR(pkg.as_mut_ptr()), &mut ifo) };
     if failed(res) {
         if ifo != ptr::null_mut() {
             unsafe {
-                FreeContextBuffer(ifo as *mut _);
+                FreeContextBuffer(ifo as *mut c_void);
             }
         }
         bail!("failed to query pkg info for Kerberos {}", format_error(res));
     }
     let max_len = unsafe { (*ifo).cbMaxToken };
     unsafe {
-        FreeContextBuffer(ifo as *mut _);
+        FreeContextBuffer(ifo as *mut c_void);
     }
     let mut buf = Vec::with_capacity(max_len as usize);
     buf.extend((0..max_len).into_iter().map(|_| 0));
@@ -154,11 +155,7 @@ fn alloc_krb5_buf() -> Result<Vec<u8>> {
 
 fn query_pkg_sizes(ctx: &mut SecHandle, sz: &mut SecPkgContext_Sizes) -> Result<()> {
     let res = unsafe {
-        QueryContextAttributesW(
-            ctx as *mut _,
-            SECPKG_ATTR_SIZES,
-            sz as *mut _ as *mut c_void,
-        )
+        QueryContextAttributesW(ctx, SECPKG_ATTR_SIZES, sz as *mut _ as *mut c_void)
     };
     if failed(res) {
         bail!("failed to query package sizes {}", format_error(res))
@@ -199,7 +196,7 @@ fn wrap_iov(
         pBuffers: buffers.as_mut_ptr(),
     };
     let flags = if !encrypt { KERB_WRAP_NO_ENCRYPT } else { 0 };
-    let res = unsafe { EncryptMessage(ctx as *mut _, flags, &mut buf_desc as *mut _, 0) };
+    let res = unsafe { EncryptMessage(ctx, flags, &mut buf_desc, 0) };
     if failed(res) {
         bail!("EncryptMessage failed {}", format_error(res))
     }
@@ -243,9 +240,7 @@ fn unwrap_iov(ctx: &mut SecHandle, len: usize, msg: &mut BytesMut) -> Result<Byt
         pBuffers: bufs.as_mut_ptr(),
     };
     let mut qop: u32 = 0;
-    let res = unsafe {
-        DecryptMessage(ctx as *mut _, &mut bufs_desc as *mut _, 0, &mut qop as *mut _)
-    };
+    let res = unsafe { DecryptMessage(ctx, &mut bufs_desc, 0, &mut qop) };
     if failed(res) {
         bail!("decrypt message failed {}", format_error(res))
     }
@@ -262,20 +257,14 @@ fn convert_lifetime(expires: i64) -> Result<Duration> {
     let mut lt = SYSTEMTIME::default();
     let mut ft = FILETIME::default();
     unsafe {
-        GetSystemTime(&mut st as *mut _);
-        if !SystemTimeToTzSpecificLocalTime(
-            ptr::null_mut(),
-            &st as *const _,
-            &mut lt as *mut _,
-        )
-        .as_bool()
-        {
+        GetSystemTime(&mut st);
+        if !SystemTimeToTzSpecificLocalTime(ptr::null_mut(), &st, &mut lt).as_bool() {
             bail!(
                 "failed to convert to local time {}",
                 format_error(GetLastError().0 as i32)
             )
         }
-        if !SystemTimeToFileTime(&lt as *const _, &mut ft as *mut _).as_bool() {
+        if !SystemTimeToFileTime(&lt, &mut ft).as_bool() {
             bail!(
                 "failed to convert current time to a filetime: {}",
                 format_error(GetLastError().0 as i32)
@@ -291,14 +280,23 @@ fn convert_lifetime(expires: i64) -> Result<Duration> {
 }
 
 #[derive(Debug)]
-pub struct PendingClientCtx(ClientCtx);
+pub(crate) struct PendingClientCtx(ClientCtx);
 
 impl PendingClientCtx {
-    pub fn finish(mut self, token: &[u8]) -> Result<ClientCtx> {
-        if self.0.do_step(Some(token))?.is_some() {
-            bail!("unexpected second token")
-        }
-        Ok(self.0)
+    pub(crate) fn step(
+        mut self,
+        token: &[u8],
+    ) -> Result<
+        Step<
+            (ClientCtx, Option<impl Deref<Target = [u8]>>),
+            (PendingClientCtx, impl Deref<Target = [u8]>),
+        >,
+    > {
+        Ok(match self.0.do_step(InputData::Subsequent(token))? {
+            None => Step::Finished((self.0, None)),
+            Some(tok) if self.0.done => Step::Finished((self.0, Some(tok))),
+            Some(tok) => Step::Continue((self, tok)),
+        })
     }
 }
 
@@ -318,7 +316,7 @@ pub struct ClientCtx {
 impl Drop for ClientCtx {
     fn drop(&mut self) {
         unsafe {
-            DeleteSecurityContext(&mut self.ctx as *mut _);
+            DeleteSecurityContext(&mut self.ctx);
         }
     }
 }
@@ -329,11 +327,18 @@ impl fmt::Debug for ClientCtx {
     }
 }
 
+#[derive(Clone, Copy, Debug)]
+enum InputData<'a> {
+    Initial(Option<&'a [u8]>),
+    Subsequent(&'a [u8]),
+}
+
 impl ClientCtx {
-    pub fn initiate(
+    pub(crate) fn new(
         flags: InitiateFlags,
         principal: Option<&str>,
         target_principal: &str,
+        cb_token: Option<&[u8]>,
     ) -> Result<(PendingClientCtx, impl Deref<Target = [u8]>)> {
         let mut ctx = ClientCtx {
             ctx: SecHandle::default(),
@@ -351,11 +356,13 @@ impl ClientCtx {
             header: BytesMut::new(),
             padding: BytesMut::new(),
         };
-        let token = ctx.do_step(None)?.ok_or_else(|| anyhow!("expected token"))?;
+        let token = ctx
+            .do_step(InputData::Initial(cb_token))?
+            .ok_or_else(|| anyhow!("expected token"))?;
         Ok((PendingClientCtx(ctx), token))
     }
 
-    fn do_step(&mut self, tok: Option<&[u8]>) -> Result<Option<BytesMut>> {
+    fn do_step(&mut self, data: InputData) -> Result<Option<BytesMut>> {
         if self.done {
             return Ok(None);
         }
@@ -365,59 +372,83 @@ impl ClientCtx {
         let mut out_buf = SecBuffer {
             cbBuffer: self.buf.len() as u32,
             BufferType: SECBUFFER_TOKEN,
-            pvBuffer: self.buf.as_mut_ptr() as *mut _,
+            pvBuffer: self.buf.as_mut_ptr() as *mut c_void,
         };
         let mut out_buf_desc = SecBufferDesc {
             ulVersion: SECBUFFER_VERSION,
             cBuffers: 1,
-            pBuffers: &mut out_buf as *mut _,
+            pBuffers: &mut out_buf,
         };
-        let mut in_buf = SecBuffer {
-            cbBuffer: tok.map(|t| t.len()).unwrap_or(0) as u32,
-            BufferType: SECBUFFER_TOKEN,
-            pvBuffer: tok
-                .map(|t| unsafe { mem::transmute::<*const u8, *mut c_void>(t.as_ptr()) })
-                .unwrap_or(ptr::null_mut()),
+        let mut cbt_buf;
+        let mut in_buf = match data {
+            InputData::Initial(None) => SecBuffer {
+                cbBuffer: 0,
+                BufferType: SECBUFFER_TOKEN,
+                pvBuffer: ptr::null_mut(),
+            },
+            InputData::Initial(Some(cb_token)) => {
+                cbt_buf = BytesMut::with_capacity(
+                    mem::size_of::<SEC_CHANNEL_BINDINGS>() + cb_token.len(),
+                );
+                let mut sec_cb = SEC_CHANNEL_BINDINGS::default();
+                sec_cb.dwApplicationDataOffset =
+                    mem::size_of::<SEC_CHANNEL_BINDINGS>() as u32;
+                sec_cb.cbApplicationDataLength = cb_token.len() as u32;
+                cbt_buf.extend(unsafe {
+                    std::slice::from_raw_parts(
+                        &sec_cb as *const SEC_CHANNEL_BINDINGS as *const u8,
+                        mem::size_of::<SEC_CHANNEL_BINDINGS>(),
+                    )
+                });
+                cbt_buf.extend(cb_token);
+                SecBuffer {
+                    cbBuffer: cbt_buf.len() as u32,
+                    BufferType: SECBUFFER_CHANNEL_BINDINGS,
+                    pvBuffer: cbt_buf.as_ptr() as *mut c_void,
+                }
+            }
+            InputData::Subsequent(tok) => SecBuffer {
+                cbBuffer: tok.len() as u32,
+                BufferType: SECBUFFER_TOKEN,
+                pvBuffer: tok.as_ptr() as *mut c_void
+            },
         };
         let mut in_buf_desc = SecBufferDesc {
             ulVersion: SECBUFFER_VERSION,
             cBuffers: 1,
-            pBuffers: &mut in_buf as *mut _,
+            pBuffers: &mut in_buf,
+        };
+        let ctx_ptr = match data {
+            InputData::Initial(_) => ptr::null_mut(),
+            InputData::Subsequent(_) => &mut self.ctx,
+        };
+        let in_buf_ptr = match data {
+            InputData::Initial(Some(_)) | InputData::Subsequent(_) => &mut in_buf_desc,
+            InputData::Initial(None) => ptr::null_mut(),
         };
-        let ctx_ptr =
-            if tok.is_none() { ptr::null_mut() } else { &mut self.ctx as *mut _ };
-        let in_buf_ptr =
-            if tok.is_some() { &mut in_buf_desc as *mut _ } else { ptr::null_mut() };
         let res = unsafe {
             InitializeSecurityContextW(
-                &mut self.cred.0 as *mut _,
+                &mut self.cred.0,
                 ctx_ptr,
-                self.target.as_mut_ptr(),
+                self.target.as_ptr(),
                 ISC_REQ_CONFIDENTIALITY | ISC_REQ_MUTUAL_AUTH,
                 0,
                 SECURITY_NATIVE_DREP,
                 in_buf_ptr,
                 0,
-                &mut self.ctx as *mut _,
-                &mut out_buf_desc as *mut _,
-                &mut self.attrs as *mut _,
-                &mut self.lifetime as *mut _,
+                &mut self.ctx,
+                &mut out_buf_desc,
+                &mut self.attrs,
+                &mut self.lifetime,
             )
         };
         if failed(res) {
             bail!("ClientCtx::step failed {}", format_error(res))
         }
-        let res = res as u32;
         if res == SEC_E_OK.0 {
             query_pkg_sizes(&mut self.ctx, &mut self.sizes)?;
             self.done = true;
         }
-        if res == SEC_I_COMPLETE_AND_CONTINUE.0 || res == SEC_I_COMPLETE_NEEDED.0 {
-            let res = unsafe { CompleteAuthToken(ctx_ptr, &mut out_buf_desc as *mut _) };
-            if failed(res) {
-                bail!("complete token failed {}", format_error(res))
-            }
-        }
         if out_buf.cbBuffer > 0 {
             Ok(Some(BytesMut::from(&self.buf[0..(out_buf.cbBuffer as usize)])))
         } else if self.done {
@@ -462,6 +493,26 @@ impl K5Ctx for ClientCtx {
     }
 }
 
+pub(crate) struct PendingServerCtx(ServerCtx);
+
+impl PendingServerCtx {
+    pub(crate) fn step(
+        mut self,
+        tok: &[u8],
+    ) -> Result<
+        Step<
+            (ServerCtx, Option<impl Deref<Target = [u8]>>),
+            (PendingServerCtx, impl Deref<Target = [u8]>),
+        >,
+    > {
+        Ok(match self.0.do_step(tok)? {
+            None => Step::Finished((self.0, None)),
+            Some(tok) if self.0.done => Step::Finished((self.0, Some(tok))),
+            Some(tok) => Step::Continue((self, tok)),
+        })
+    }
+}
+
 pub struct ServerCtx {
     ctx: SecHandle,
     cred: Cred,
@@ -477,7 +528,7 @@ pub struct ServerCtx {
 impl Drop for ServerCtx {
     fn drop(&mut self) {
         unsafe {
-            DeleteSecurityContext(&mut self.ctx as *mut _);
+            DeleteSecurityContext(&mut self.ctx);
         }
     }
 }
@@ -489,12 +540,11 @@ impl fmt::Debug for ServerCtx {
 }
 
 impl ServerCtx {
-    pub fn accept(
+    pub(crate) fn new(
         flags: AcceptFlags,
         principal: Option<&str>,
-        token: &[u8],
-    ) -> Result<(Self, impl Deref<Target = [u8]>)> {
-        let mut ctx = ServerCtx {
+    ) -> Result<PendingServerCtx> {
+        Ok(PendingServerCtx(ServerCtx {
             ctx: SecHandle::default(),
             cred: Cred::acquire(
                 flags.contains(AcceptFlags::NEGOTIATE_TOKEN),
@@ -508,13 +558,10 @@ impl ServerCtx {
             sizes: SecPkgContext_Sizes::default(),
             header: BytesMut::new(),
             padding: BytesMut::new(),
-        };
-        let token = ctx.do_step(Some(token))?.ok_or_else(|| anyhow!("expected token"))?;
-        Ok((ctx, token))
+        }))
     }
 
-    fn do_step(&mut self, tok: Option<&[u8]>) -> Result<Option<BytesMut>> {
-        let tok = tok.ok_or_else(|| anyhow!("no token supplied"))?;
+    fn do_step(&mut self, tok: &[u8]) -> Result<Option<BytesMut>> {
         if self.done {
             return Ok(None);
         }
@@ -524,12 +571,12 @@ impl ServerCtx {
         let mut out_buf = SecBuffer {
             cbBuffer: self.buf.len() as u32,
             BufferType: SECBUFFER_TOKEN,
-            pvBuffer: self.buf.as_mut_ptr() as *mut _,
+            pvBuffer: self.buf.as_mut_ptr() as *mut c_void,
         };
         let mut out_buf_desc = SecBufferDesc {
             ulVersion: SECBUFFER_VERSION,
             cBuffers: 1,
-            pBuffers: &mut out_buf as *mut _,
+            pBuffers: &mut out_buf,
         };
         let mut in_buf = SecBuffer {
             cbBuffer: tok.len() as u32,
@@ -539,44 +586,37 @@ impl ServerCtx {
         let mut in_buf_desc = SecBufferDesc {
             ulVersion: SECBUFFER_VERSION,
             cBuffers: 1,
-            pBuffers: &mut in_buf as *mut _,
+            pBuffers: &mut in_buf,
         };
         let dfsh = SecHandle::default();
         let ctx_ptr =
             if self.ctx.dwLower == dfsh.dwLower && self.ctx.dwUpper == dfsh.dwUpper {
                 ptr::null_mut()
             } else {
-                &mut self.ctx as *mut _
+                &mut self.ctx
             };
         let res = unsafe {
             AcceptSecurityContext(
-                &mut self.cred.0 as *mut _,
+                &mut self.cred.0,
                 ctx_ptr,
-                &mut in_buf_desc as *mut _,
+                &mut in_buf_desc,
                 ACCEPT_SECURITY_CONTEXT_CONTEXT_REQ(
                     ISC_REQ_CONFIDENTIALITY | ISC_REQ_MUTUAL_AUTH,
                 ),
                 SECURITY_NATIVE_DREP,
-                &mut self.ctx as *mut _,
-                &mut out_buf_desc as *mut _,
-                &mut self.attrs as *mut _,
-                &mut self.lifetime as *mut _,
+                &mut self.ctx,
+                &mut out_buf_desc,
+                &mut self.attrs,
+                &mut self.lifetime,
             )
         };
         if failed(res) {
             bail!("ServerCtx::step failed {}", format_error(res));
         }
-        let res = res as u32;
         if res == SEC_E_OK.0 {
             query_pkg_sizes(&mut self.ctx, &mut self.sizes)?;
             self.done = true;
         }
-        if res == SEC_I_COMPLETE_AND_CONTINUE.0 || res == SEC_I_COMPLETE_NEEDED.0 {
-            let res = unsafe { CompleteAuthToken(ctx_ptr, &mut out_buf_desc as *mut _) };
-            if failed(res) {
-                bail!("complete token failed {}", format_error(res))
-            }
-        }
         if out_buf.cbBuffer > 0 {
             Ok(Some(BytesMut::from(&self.buf[0..(out_buf.cbBuffer as usize)])))
         } else if self.done {
@@ -626,7 +666,7 @@ impl K5ServerCtx for ServerCtx {
         let mut names = SecPkgContext_NativeNamesW::default();
         unsafe {
             let res = QueryContextAttributesW(
-                &mut self.ctx as *mut _,
+                &mut self.ctx,
                 SECPKG_ATTR_NATIVE_NAMES,
                 &mut names as *mut _ as *mut c_void,
             );
